"""vfolder-and-kernel

Revision ID: 7ea324d0535b
Revises: 5de06da3c2b5
Create Date: 2017-08-08 16:25:59.553570

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "7ea324d0535b"
down_revision = "5de06da3c2b5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "agents",
        sa.Column("id", sa.String(length=64), nullable=False),
        sa.Column("mem_slots", sa.Integer(), nullable=False),
        sa.Column("cpu_slots", sa.Integer(), nullable=False),
        sa.Column("gpu_slots", sa.Integer(), nullable=False),
        sa.Column("used_mem_slots", sa.Integer(), nullable=False),
        sa.Column("used_cpu_slots", sa.Integer(), nullable=False),
        sa.Column("used_gpu_slots", sa.Integer(), nullable=False),
        sa.Column("addr", sa.String(length=128), nullable=False),
        sa.Column(
            "first_contact",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_agents")),
    )
    op.create_table(
        "vfolders",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("host", sa.String(length=128), nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("max_files", sa.Integer(), nullable=True),
        sa.Column("max_size", sa.Integer(), nullable=True),
        sa.Column("num_files", sa.Integer(), nullable=True),
        sa.Column("cur_size", sa.Integer(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column("last_used", sa.DateTime(timezone=True), nullable=True),
        sa.Column("belongs_to", sa.String(length=20), nullable=False),
        sa.ForeignKeyConstraint(
            ["belongs_to"], ["keypairs.access_key"], name=op.f("fk_vfolders_belongs_to_keypairs")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_vfolders")),
    )
    op.create_table(
        "vfolder_attachment",
        sa.Column("vfolder", GUID(), nullable=False),
        sa.Column("kernel", GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["kernel"], ["kernels.sess_id"], name=op.f("fk_vfolder_attachment_kernel_kernels")
        ),
        sa.ForeignKeyConstraint(
            ["vfolder"], ["vfolders.id"], name=op.f("fk_vfolder_attachment_vfolder_vfolders")
        ),
        sa.PrimaryKeyConstraint("vfolder", "kernel", name=op.f("pk_vfolder_attachment")),
    )
    op.drop_table("usage")
    op.add_column("kernels", sa.Column("agent", sa.String(length=64), nullable=True))
    op.add_column("kernels", sa.Column("allocated_cores", sa.ARRAY(sa.Integer()), nullable=True))
    op.add_column("kernels", sa.Column("cpu_used", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("cur_mem_bytes", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("io_read_bytes", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("io_write_bytes", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("max_mem_bytes", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("net_rx_bytes", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("net_tx_bytes", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("num_queries", sa.BigInteger(), nullable=True))
    op.add_column("kernels", sa.Column("status_info", sa.Unicode(), nullable=True))
    op.create_index(op.f("ix_kernels_created_at"), "kernels", ["created_at"], unique=False)
    op.create_index(op.f("ix_kernels_status"), "kernels", ["status"], unique=False)
    op.create_index(op.f("ix_kernels_terminated_at"), "kernels", ["terminated_at"], unique=False)
    op.create_foreign_key(op.f("fk_kernels_agent_agents"), "kernels", "agents", ["agent"], ["id"])
    op.drop_column("kernels", "agent_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "kernels", sa.Column("agent_id", sa.VARCHAR(), autoincrement=False, nullable=True)
    )
    op.drop_constraint(op.f("fk_kernels_agent_agents"), "kernels", type_="foreignkey")
    op.drop_index(op.f("ix_kernels_terminated_at"), table_name="kernels")
    op.drop_index(op.f("ix_kernels_status"), table_name="kernels")
    op.drop_index(op.f("ix_kernels_created_at"), table_name="kernels")
    op.drop_column("kernels", "status_info")
    op.drop_column("kernels", "num_queries")
    op.drop_column("kernels", "net_tx_bytes")
    op.drop_column("kernels", "net_rx_bytes")
    op.drop_column("kernels", "max_mem_bytes")
    op.drop_column("kernels", "io_write_bytes")
    op.drop_column("kernels", "io_read_bytes")
    op.drop_column("kernels", "cur_mem_bytes")
    op.drop_column("kernels", "cpu_used")
    op.drop_column("kernels", "allocated_cores")
    op.drop_column("kernels", "agent")
    op.create_table(
        "usage",
        sa.Column(
            "id",
            postgresql.UUID(),
            server_default=sa.text("uuid_generate_v4()"),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column("access_key_id", sa.VARCHAR(length=20), autoincrement=False, nullable=True),
        sa.Column("kernel_type", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column("kernel_id", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "started_at", postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=True
        ),
        sa.Column(
            "terminated_at", postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=True
        ),
        sa.Column(
            "cpu_used",
            sa.INTEGER(),
            server_default=sa.text("0"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "mem_used",
            sa.INTEGER(),
            server_default=sa.text("0"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "io_used", sa.INTEGER(), server_default=sa.text("0"), autoincrement=False, nullable=True
        ),
        sa.Column(
            "net_used",
            sa.INTEGER(),
            server_default=sa.text("0"),
            autoincrement=False,
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["access_key_id"], ["keypairs.access_key"], name="fk_usage_access_key_id_keypairs"
        ),
        sa.PrimaryKeyConstraint("id", name="pk_usage"),
    )
    op.drop_table("vfolder_attachment")
    op.drop_table("vfolders")
    op.drop_table("agents")
    # ### end Alembic commands ###



import contextlib
import logging
from pprint import pprint
from typing import AsyncIterator

import sqlalchemy as sa
from tabulate import tabulate

from ai.backend.common.docker import ImageRef
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.manager.models.image import ImageAliasRow, ImageRow
from ai.backend.manager.models.image import rescan_images as rescan_images_func
from ai.backend.manager.models.utils import connect_database

log = BraceStyleAdapter(logging.getLogger(__name__))


@contextlib.asynccontextmanager
async def etcd_ctx(cli_ctx) -> AsyncIterator[AsyncEtcd]:
    local_config = cli_ctx.local_config
    creds = None
    if local_config["etcd"]["user"]:
        creds = {
            "user": local_config["etcd"]["user"],
            "password": local_config["etcd"]["password"],
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: "",
        # TODO: provide a way to specify other scope prefixes
    }
    etcd = AsyncEtcd(
        local_config["etcd"]["addr"],
        local_config["etcd"]["namespace"],
        scope_prefix_map,
        credentials=creds,
    )
    try:
        yield etcd
    finally:
        await etcd.close()


async def list_images(cli_ctx, short, installed):
    async with connect_database(cli_ctx.local_config) as db:
        async with db.begin_readonly_session() as session:
            displayed_items = []
            try:
                items = await ImageRow.list(session)
                # NOTE: installed/installed_agents fields are no longer provided in CLI,
                #       until we finish the epic refactoring of image metadata db.
                for item in items:
                    if installed and not item.installed:
                        continue
                    if short:
                        displayed_items.append((item.image_ref.canonical, item.config_digest))
                    else:
                        pprint(item)
                if short:
                    print(tabulate(displayed_items, tablefmt="plain"))
            except Exception:
                log.exception("An error occurred.")


async def inspect_image(cli_ctx, canonical_or_alias, architecture):
    async with connect_database(cli_ctx.local_config) as db:
        async with db.begin_readonly_session() as session:
            try:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageRef(canonical_or_alias, ["*"], architecture),
                        canonical_or_alias,
                    ],
                )
                pprint(await image_row.inspect())
            except UnknownImageReference:
                log.exception("Image not found.")
            except Exception:
                log.exception("An error occurred.")


async def forget_image(cli_ctx, canonical_or_alias, architecture):
    async with connect_database(cli_ctx.local_config) as db:
        async with db.begin_session() as session:
            try:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageRef(canonical_or_alias, ["*"], architecture),
                        canonical_or_alias,
                    ],
                )
                await session.delete(image_row)
            except UnknownImageReference:
                log.exception("Image not found.")
            except Exception:
                log.exception("An error occurred.")


async def set_image_resource_limit(
    cli_ctx,
    canonical_or_alias,
    slot_type,
    range_value,
    architecture,
):
    async with connect_database(cli_ctx.local_config) as db:
        async with db.begin_session() as session:
            try:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageRef(canonical_or_alias, ["*"], architecture),
                        canonical_or_alias,
                    ],
                )
                await image_row.set_resource_limit(slot_type, range_value)
            except UnknownImageReference:
                log.exception("Image not found.")
            except Exception:
                log.exception("An error occurred.")


async def rescan_images(cli_ctx, registry):
    async with connect_database(cli_ctx.local_config) as db:
        async with etcd_ctx(cli_ctx) as etcd:
            try:
                await rescan_images_func(etcd, db, registry=registry)
            except Exception:
                log.exception("An error occurred.")


async def alias(cli_ctx, alias, target, architecture):
    async with connect_database(cli_ctx.local_config) as db:
        async with db.begin_session() as session:
            try:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageRef(target, ["*"], architecture),
                    ],
                )
                await ImageAliasRow.create(session, alias, image_row)
            except UnknownImageReference:
                log.exception("Image not found.")
            except Exception:
                log.exception("An error occurred.")


async def dealias(cli_ctx, alias):
    async with connect_database(cli_ctx.local_config) as db:
        async with db.begin_session() as session:
            alias_row = await session.scalar(
                sa.select(ImageAliasRow).where(ImageAliasRow.alias == alias),
            )
            if alias_row is None:
                log.exception("Alias not found.")
                return
            await session.delete(alias_row)



"""add vfolder_permissions table

Revision ID: 185852ff9872
Revises: 1fa6a31ea8e3
Create Date: 2018-07-05 16:02:05.225094

"""
import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "185852ff9872"
down_revision = "1fa6a31ea8e3"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "vfolder_permissions",
        sa.Column("permission", sa.String(length=2), nullable=True),
        sa.Column("vfolder", GUID(), nullable=False),
        sa.Column("access_key", sa.String(length=20), nullable=False),
        sa.ForeignKeyConstraint(
            ["access_key"],
            ["keypairs.access_key"],
            name=op.f("fk_vfolder_permissions_access_key_keypairs"),
        ),
        sa.ForeignKeyConstraint(
            ["vfolder"],
            ["vfolders.id"],
            name=op.f("fk_vfolder_permissions_vfolder_vfolders"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("vfolder_permissions")
    # ### end Alembic commands ###



from __future__ import annotations

import logging

import aiohttp_cors
from aiohttp import web

from ai.backend.common.logging import BraceStyleAdapter

from ..models.acl import get_all_permissions
from .auth import auth_required
from .manager import ALL_ALLOWED, server_status_required

log = BraceStyleAdapter(logging.getLogger(__name__))


@auth_required
@server_status_required(ALL_ALLOWED)
async def get_permission(request: web.Request) -> web.Response:
    access_key = request["keypair"]["access_key"]
    log.info("GET_PERMISSION (ak:{})", access_key)

    return web.json_response(get_all_permissions(), status=200)


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(default_cors_options):
    app = web.Application()
    app["prefix"] = "acl"
    app["api_versions"] = (4,)
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", get_permission))
    return app, []



import datetime
import json
import logging
import uuid
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Tuple

import aiohttp_cors
import sqlalchemy as sa
import trafaret as t
import yaml
from aiohttp import web

from ai.backend.common import validators as tx
from ai.backend.common.logging import BraceStyleAdapter

from ..models import TemplateType, groups, session_templates, users
from ..models.session_template import check_task_template
from .auth import auth_required
from .exceptions import InvalidAPIParameters, TaskTemplateNotFound
from .manager import READ_ALLOWED, server_status_required
from .session import _query_userinfo
from .types import CORSOptions, Iterable, WebMiddleware
from .utils import check_api_params, get_access_key_scopes

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["group", "groupName", "group_name"], default="default"): t.String,
            tx.AliasedKey(["domain", "domainName", "domain_name"], default="default"): t.String,
            t.Key("owner_access_key", default=None): t.Null | t.String,
            t.Key("payload"): t.String,
        },
    )
)
async def create(request: web.Request, params: Any) -> web.Response:
    if params["domain"] is None:
        params["domain"] = request["user"]["domain_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        "SESSION_TEMPLATE.CREATE (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        user_uuid, group_id, _ = await _query_userinfo(request, params, conn)
        log.debug("Params: {0}", params)
        try:
            body = json.loads(params["payload"])
        except json.JSONDecodeError:
            try:
                body = yaml.safe_load(params["payload"])
            except (yaml.YAMLError, yaml.MarkedYAMLError):
                raise InvalidAPIParameters("Malformed payload")
        for st in body["session_templates"]:
            template_data = check_task_template(st["template"])
            template_id = uuid.uuid4().hex
            resp = {
                "id": template_id,
                "user": user_uuid.hex,
            }
            name = st["name"] if "name" in st else template_data["metadata"]["name"]
            if "group_id" in st:
                group_id = st["group_id"]
            if "user_uuid" in st:
                user_uuid = st["user_uuid"]
            query = session_templates.insert().values(
                {
                    "id": template_id,
                    "created_at": datetime.datetime.now(),
                    "domain_name": params["domain"],
                    "group_id": group_id,
                    "user_uuid": user_uuid,
                    "name": name,
                    "template": template_data,
                    "type": TemplateType.TASK,
                }
            )
            result = await conn.execute(query)
            assert result.rowcount == 1
    return web.json_response(resp)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("all", default=False): t.ToBool,
            tx.AliasedKey(["group_id", "groupId"], default=None): tx.UUID | t.String | t.Null,
        }
    ),
)
async def list_template(request: web.Request, params: Any) -> web.Response:
    resp = []
    access_key = request["keypair"]["access_key"]
    domain_name = request["user"]["domain_name"]
    user_uuid = request["user"]["uuid"]
    log.info("SESSION_TEMPLATE.LIST (ak:{})", access_key)
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        entries: List[Mapping[str, Any]]
        j = session_templates.join(
            users, session_templates.c.user_uuid == users.c.uuid, isouter=True
        ).join(groups, session_templates.c.group_id == groups.c.id, isouter=True)
        query = (
            sa.select([session_templates, users.c.email, groups.c.name], use_labels=True)
            .select_from(j)
            .where(
                (session_templates.c.is_active) & (session_templates.c.type == TemplateType.TASK),
            )
        )
        result = await conn.execute(query)
        entries = []
        for row in result.fetchall():
            is_owner = True if row.session_templates_user_uuid == user_uuid else False
            entries.append(
                {
                    "name": row.session_templates_name,
                    "id": row.session_templates_id,
                    "created_at": row.session_templates_created_at,
                    "is_owner": is_owner,
                    "user": (
                        str(row.session_templates_user_uuid)
                        if row.session_templates_user_uuid
                        else None
                    ),
                    "group": (
                        str(row.session_templates_group_id)
                        if row.session_templates_group_id
                        else None
                    ),
                    "user_email": row.users_email,
                    "group_name": row.groups_name,
                    "domain_name": domain_name,
                    "type": row.session_templates_type,
                    "template": row.session_templates_template,
                }
            )
        for entry in entries:
            resp.append(
                {
                    "name": entry["name"],
                    "id": entry["id"].hex,
                    "created_at": str(entry["created_at"]),
                    "is_owner": entry["is_owner"],
                    "user": str(entry["user"]),
                    "group": str(entry["group"]),
                    "user_email": entry["user_email"],
                    "group_name": entry["group_name"],
                    "domain_name": domain_name,
                    "type": entry["type"],
                    "template": entry["template"],
                }
            )
        return web.json_response(resp)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("format", default="json"): t.Null | t.Enum("yaml", "json"),
            t.Key("owner_access_key", default=None): t.Null | t.String,
        }
    ),
)
async def get(request: web.Request, params: Any) -> web.Response:
    if params["format"] not in ["yaml", "json"]:
        raise InvalidAPIParameters('format should be "yaml" or "json"')
    resp: Dict[str, Any] = {}
    domain_name = request["user"]["domain_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        "SESSION_TEMPLATE.GET (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )
    template_id = request.match_info["template_id"]
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        query = (
            sa.select(
                [
                    session_templates.c.template,
                    session_templates.c.name,
                    session_templates.c.user_uuid,
                    session_templates.c.group_id,
                ]
            )
            .select_from(session_templates)
            .where(
                (session_templates.c.id == template_id)
                & (session_templates.c.is_active)
                & (session_templates.c.type == TemplateType.TASK),
            )
        )
        result = await conn.execute(query)
        for row in result.fetchall():
            resp.update(
                {
                    "template": row.template,
                    "name": row.name,
                    "user_uuid": str(row.user_uuid),
                    "group_id": str(row.group_id),
                    "domain_name": domain_name,
                }
            )
        if isinstance(resp, str):
            resp = json.loads(resp)
        else:
            resp = json.loads(json.dumps(resp))
        return web.json_response(resp)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["group", "groupName", "group_name"], default="default"): t.String,
            tx.AliasedKey(["domain", "domainName", "domain_name"], default="default"): t.String,
            t.Key("payload"): t.String,
            t.Key("owner_access_key", default=None): t.Null | t.String,
        }
    ),
)
async def put(request: web.Request, params: Any) -> web.Response:
    if params["domain"] is None:
        params["domain"] = request["user"]["domain_name"]
    template_id = request.match_info["template_id"]

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        "SESSION_TEMPLATE.PUT (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        user_uuid, group_id, _ = await _query_userinfo(request, params, conn)
        query = (
            sa.select([session_templates.c.id])
            .select_from(session_templates)
            .where(
                (session_templates.c.id == template_id)
                & (session_templates.c.is_active)
                & (session_templates.c.type == TemplateType.TASK),
            )
        )
        result = await conn.scalar(query)
        if not result:
            raise TaskTemplateNotFound
        try:
            body = json.loads(params["payload"])
        except json.JSONDecodeError:
            body = yaml.safe_load(params["payload"])
        except (yaml.YAMLError, yaml.MarkedYAMLError):
            raise InvalidAPIParameters("Malformed payload")
        for st in body["session_templates"]:
            template_data = check_task_template(st["template"])
            name = st["name"] if "name" in st else template_data["metadata"]["name"]
            if "group_id" in st:
                group_id = st["group_id"]
            if "user_uuid" in st:
                user_uuid = st["user_uuid"]
            query = (
                sa.update(session_templates)
                .values(
                    {
                        "group_id": group_id,
                        "user_uuid": user_uuid,
                        "name": name,
                        "template": template_data,
                    }
                )
                .where((session_templates.c.id == template_id))
            )
            result = await conn.execute(query)
            assert result.rowcount == 1
        return web.json_response({"success": True})


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("owner_access_key", default=None): t.Null | t.String,
        }
    ),
)
async def delete(request: web.Request, params: Any) -> web.Response:
    template_id = request.match_info["template_id"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        "SESSION_TEMPLATE.DELETE (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        query = (
            sa.select([session_templates.c.id])
            .select_from(session_templates)
            .where(
                (session_templates.c.id == template_id)
                & (session_templates.c.is_active)
                & (session_templates.c.type == TemplateType.TASK),
            )
        )
        result = await conn.scalar(query)
        if not result:
            raise TaskTemplateNotFound
        query = (
            sa.update(session_templates)
            .values(is_active=False)
            .where((session_templates.c.id == template_id))
        )
        result = await conn.execute(query)
        assert result.rowcount == 1

        return web.json_response({"success": True})


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["api_versions"] = (4, 5)
    app["prefix"] = "template/session"
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("POST", "", create))
    cors.add(app.router.add_route("GET", "", list_template))
    template_resource = cors.add(app.router.add_resource(r"/{template_id}"))
    cors.add(template_resource.add_route("GET", get))
    cors.add(template_resource.add_route("PUT", put))
    cors.add(template_resource.add_route("DELETE", delete))

    return app, []
